<?php

namespace Drupal\twig_ui\Entity;

use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Render\Element\Checkboxes;
use Drupal\twig_ui\TemplateManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form controller for the Twig UI template edit/add forms.
 */
class TwigTemplateForm extends EntityForm {

  /**
   * The Twig UI Template Manager service.
   *
   * @var \Drupal\twig_ui\TemplateManager
   */
  protected $templateManager;

  /**
   * The entityTypeManager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The messenger service.
   *
   * @var \Drupal\Core\Messenger\Messenger
   */
  protected $messenger;

  /**
   * Constructs a TwigTemplate Form object.
   *
   * @param \Drupal\twig_ui\TemplateManager $template_manager
   *   The Twig UI Template Manager service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entityTypeManager.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   */
  public function __construct(TemplateManager $template_manager, EntityTypeManagerInterface $entityTypeManager, ModuleHandlerInterface $module_handler, MessengerInterface $messenger) {
    $this->templateManager = $template_manager;
    $this->entityTypeManager = $entityTypeManager;
    $this->moduleHandler = $module_handler;
    $this->messenger = $messenger;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('twig_ui.template_manager'),
      $container->get('entity_type.manager'),
      $container->get('module_handler'),
      $container->get('messenger')
    );
  }

  /**
   * The add/edit Twig Template entity form.
   *
   * @param array $form
   *   A nested array form elements comprising the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return array
   *   The array containing the complete form.
   */
  public function form(array $form, FormStateInterface $form_state) {
    if ($this->operation == 'edit') {
      $form['#title'] = $this->t('<em>Edit Twig template provider</em> @label', ['@label' => $this->entity->label()]);
    }

    $entity = $this->entity;

    $form['label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Template name'),
      '#default_value' => $entity->label(),
      '#required' => TRUE,
    ];

    $form['id'] = [
      '#type' => 'machine_name',
      '#title' => $this->t('Machine name'),
      '#default_value' => $entity->id(),
      '#machine_name' => [
        'exists' => [$this, 'exist'],
      ],
    ];

    $form['theme_suggestion'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Theme suggestion'),
      '#description' => $this->t('See <a href="@url" target="_blank">Twig Template naming conventions</a>.', ['@url' => 'https://www.drupal.org/docs/theming-drupal/twig-in-drupal/twig-template-naming-conventions']),
      '#default_value' => $entity->get('theme_suggestion'),
      '#required' => TRUE,
    ];

    $form['template_code'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Template code'),
      '#default_value' => $entity->get('template_code'),
      '#required' => TRUE,
    ];
    if ($this->moduleHandler->moduleExists('codemirror_editor')) {
      $form['template_code']['#type'] = 'codemirror';
      $form['template_code']['#codemirror'] = [
        'mode' => 'text/x-twig',
         // 'buttons' => [], // @TODO Allow configuration of buttons.
      ];
    }

    $options = [];
    $list = $this->templateManager->getActiveThemes();
    foreach ($list as $theme_name => $theme) {
      $options[$theme_name] = $theme->info['name'];
    }

    $form['themes'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Themes'),
      '#description' => $this->t('Themes for which this template should be used.'),
      '#options' => $options,
      '#default_value' => $entity->get('themes') ?: [],
    ];

    return parent::form($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    // Run 'themes' values through Checkboxes::getCheckedCheckboxes()
    // to remove unchecked values.
    if (!empty($form_state->getValue('themes'))) {
      $form_state->setValue('themes', Checkboxes::getCheckedCheckboxes($form_state->getValue('themes')));
    }

    $values = $form_state->getValues();

    // Loop through submitted themes and check if the theme suggestion is
    // already registered for the selected themes.
    foreach ($values['themes'] as $theme) {
      if ($template_id = $this->templateManager->templateExists($values['theme_suggestion'], $theme)) {
        if ($template_id != $this->entity->id()) {
          $theme_label = $this->templateManager->getActiveThemes()[$theme]->info['name'];

          $template = $this->templateManager->getTemplate($template_id);
          $template_url = $template->toLink(NULL, 'edit-form')->getUrl()->toString();
          $form_state->setError(
            $form['themes'],
            $this->t(
              'The <em>@suggestion</em> theme suggestion is already registered for the <em>@theme</em> theme by the <em><a href="@template_url">@template_label</a></em> Twig UI template.', [
                '@suggestion' => $values['theme_suggestion'],
                '@theme' => $theme_label,
                '@template_label' => $template->label(),
                '@template_url' => $template_url,
              ]
            ),
          );
        }
      }
    }
    parent::validateForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $this->entity->save();
    $form_state->setRedirect('entity.twig_template.collection');
  }

  /**
   * Helper method to check whether a Twig UI template config entity exists.
   *
   * @param string $id
   *   A machine name.
   *
   * @return bool
   *   Whether or not the machine name already exists.
   */
  public function exist($id) {
    // The 'add' namespace is reserved.
    if ($id == 'add') {
      return TRUE;
    }
    $entity = $this->entityTypeManager->getStorage('twig_template')->getQuery()
      ->condition('id', $id)
      ->execute();
    return (bool) $entity;
  }

}
