<?php

namespace Drupal\twig_ui;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ThemeHandlerInterface;
use Drupal\Core\File\FileSystem;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\StreamWrapper\StreamWrapperManagerInterface;
use Drupal\twig_ui\Entity\TwigTemplate;

/**
 * The TemplateManager class.
 */
class TemplateManager {

  /**
   * The base directory path for Twig UI template files.
   *
   * This constant must use a stream wrapper (e.g. private://).
   */
  const DIRECTORY_PATH = 'public://twig_ui';

  /**
   * The entityTypeManager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The theme handler.
   *
   * @var \Drupal\Core\Extension\ThemeHandlerInterface
   */
  protected $themeHandler;

  /**
   * An interface for helpers that operate on files and stream wrappers.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * The stream wrapper manager service.
   *
   * @var \Drupal\Core\StreamWrapper\StreamWrapperManagerInterface
   */
  protected $streamWrapperManager;

  /**
   * Constructs a TemplateManager object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entityTypeManager.
   * @param \Drupal\Core\Extension\ThemeHandlerInterface $theme_handler
   *   The theme handler.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   An interface for helpers that operate on files and stream wrappers.
   * @param \Drupal\Core\StreamWrapper\StreamWrapperManagerInterface $stream_wrapper_manager
   *   The stream wrapper manager service.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, ThemeHandlerInterface $theme_handler, FileSystemInterface $file_system, StreamWrapperManagerInterface $stream_wrapper_manager) {
    $this->entityTypeManager = $entity_type_manager;
    $this->themeHandler = $theme_handler;
    $this->fileSystem = $file_system;
    $this->streamWrapperManager = $stream_wrapper_manager;
  }

  /**
   * Retrieves a listing of all Twig UI templates.
   *
   * @return array
   *   An array of TwigTemplate objects.
   */
  public function getTemplates() {
    $templates = $this->entityTypeManager
      ->getStorage('twig_template')
      ->loadMultiple();

    return $templates;
  }

  /**
   * Retrieves all Twig UI templates registered for a given theme.
   *
   * @param string $theme
   *   The machine name of a theme.
   *
   * @return mixed
   *   An array of TwigTemplate objects; NULL if none are found.
   */
  public function getTemplatesByTheme($theme) {
    $query = $this->entityTypeManager
      ->getStorage('twig_template')
      ->getQuery('AND');
    // See
    // https://www.drupal.org/project/drupal/issues/2248567#comment-13080439.
    $query->condition('themes.*', $theme);
    $ids = $query->execute();

    $templates = $this->entityTypeManager
      ->getStorage('twig_template')
      ->loadMultiple($ids);

    return (!empty($templates)) ? $templates : NULL;
  }

  /**
   * Retrieves a Twig UI template by machine name.
   *
   * @param string $id
   *   The machine name of a Twig UI template.
   *
   * @return mixed
   *   A \Drupal\twig_ui\Entity\TwigTemplate object is a template is found.
   *   NULL if no template is found.
   */
  public function getTemplate($id) {
    $templates = $this->entityTypeManager
      ->getStorage('twig_template')
      ->loadByProperties(['id' => $id]);

    return (is_array($templates)) ? array_shift($templates) : NULL;
  }

  /**
   * Checks if a Twig UI template exists for a given theme.
   *
   * @param string $suggestion
   *   A theme suggestion (e.g. template name with underscores).
   * @param string $theme
   *   The machine name of a theme.
   *
   * @return mixed
   *   Machine name of Twig UI template; FALSE if templates does not exist.
   */
  public function templateExists($suggestion, $theme) {
    $query = $this->entityTypeManager
      ->getStorage('twig_template')
      ->getQuery('AND');
    $query->condition('theme_suggestion', $suggestion);
    // See
    // https://www.drupal.org/project/drupal/issues/2248567#comment-13080439.
    $query->condition('themes.*', $theme);
    $templates = $query->execute();

    return (!empty($templates)) ? array_shift($templates) : FALSE;
  }

  /**
   * Gets the active themes.
   *
   * @return array
   *   An array of Drupal\Core\Extension\Extension objects.
   */
  public function getActiveThemes() {
    return $this->themeHandler->listInfo();
  }

  /**
   * Syncs Twig UI template entities with template files in the file system.
   *
   * @param \Drupal\twig_ui\Entity\TwigTemplate $template
   *   A Twig UI template.
   */
  public function syncTemplateFiles(TwigTemplate $template) {
    // Process template files for existing Twig UI template.
    if (isset($template->original)) {
      $original = $template->original;
      $original_theme_suggestion = $original->get('theme_suggestion');
      $theme_suggestion = $template->get('theme_suggestion');
      $original_code = $original->get('template_code');
      $code = $template->get('template_code');
      $original_themes = $original->get('themes');
      $themes = $template->get('themes');

      // If the theme_suggestion has changed or if the template code has
      // changed, then delete all files for this Twig UI template and write
      // them again.
      if ($original_theme_suggestion != $theme_suggestion
        || $original_code != $code
        ) {
        $this->deleteTemplateFiles($original);

        foreach ($template->get('themes') as $theme) {
          $this->writeTemplateFile($template, $theme);
        }
      }
      // Otherwise, check if the designated themes has changed for this
      // Twig UI template.
      elseif ($original_themes != $themes) {
        // Remove template files.
        $remove_theme_templates = array_diff($original_themes, $themes);
        foreach ($remove_theme_templates as $theme) {
          $this->deleteTemplateFile($template, $theme);
        }

        // Add template files.
        $add_theme_templates = array_diff($themes, $original_themes);
        foreach ($add_theme_templates as $theme) {
          $this->writeTemplateFile($template, $theme);
        }
      }
    }
    // Process template files for new Twig UI template.
    else {
      foreach ($template->get('themes') as $theme) {
        $this->writeTemplateFile($template, $theme);
      }
    }
  }

  /**
   * Deletes all Twig UI template files.
   *
   * @param \Drupal\twig_ui\Entity\TwigTemplate $template
   *   A Twig UI template.
   */
  public function deleteTemplateFiles(TwigTemplate $template) {
    foreach ($template->get('themes') as $theme) {
      $this->deleteTemplateFile($template, $theme);
    }
  }

  /**
   * Writes a template file to the filesystem.
   *
   * @param \Drupal\twig_ui\Entity\TwigTemplate $template
   *   A Twig UI template.
   * @param string $theme
   *   The machine name of a theme.
   */
  protected function writeTemplateFile(TwigTemplate $template, $theme) {
    $template_code = $template->get('template_code');
    $dir_path = self::getDirectoryPathByTheme($theme);
    $this->fileSystem->prepareDirectory($dir_path, FileSystem::CREATE_DIRECTORY);
    $this->fileSystem->saveData($template_code, $this->getTemplatePath($template, $theme), FileSystem::EXISTS_REPLACE);
  }

  /**
   * Deletes a template file from the filesystem.
   *
   * @param \Drupal\twig_ui\Entity\TwigTemplate $template
   *   A Twig UI template.
   * @param string $theme
   *   The machine name of a theme.
   */
  protected function deleteTemplateFile(TwigTemplate $template, $theme) {
    $this->fileSystem->delete($this->getTemplatePath($template, $theme));
  }

  /**
   * Returns the directory for a given theme.
   *
   * @param string $theme
   *   The machine name of a theme.
   * @param bool $stream_wrapper
   *   Whether or not the path should be returned in a stream wrapper.
   *
   * @return string
   *   The directory for a given theme.
   */
  public function getDirectoryPathByTheme($theme, $stream_wrapper = TRUE) {
    $path = self::DIRECTORY_PATH . '/' . $theme;

    // Returns path with stream wrapper (e.g. wrapper://twig_ui/my_theme).
    if ($stream_wrapper) {
      return $path;
    }
    // Returns path as a relative path from the docroot
    // (e.g. sites/default/files/twig_ui/my_theme).
    else {
      $stream_wrapper = $this->streamWrapperManager->getViaUri($path);
      list($scheme) = explode('://', $path, 2);
      return str_replace($scheme . '://', $stream_wrapper->basePath() . '/', $path);
    }
  }

  /**
   * Generates the template file's path.
   *
   * @param Drupal\twig_ui\Entity\TwigTemplate $template
   *   A Twig UI template.
   * @param string $theme
   *   The machine name of a theme.
   *
   * @return string
   *   The template file's path, including the filename.
   */
  public function getTemplatePath(TwigTemplate $template, $theme) {
    return self::getDirectoryPathByTheme($theme) . '/' . $this->getTemplateFileName($template);
  }

  /**
   * Generates the template file's name.
   *
   * @param Drupal\twig_ui\Entity\TwigTemplate $template
   *   A Twig UI template.
   * @param bool $extension
   *   Whether or not the file extension (.html.twig) be included.
   *
   * @return string
   *   The template file name.
   */
  public function getTemplateFileName(TwigTemplate $template, $extension = TRUE) {
    $extension_string = ($extension) ? '.html.twig' : '';
    return str_replace('_', '-', $template->get('theme_suggestion') . $extension_string);
  }

  /**
   * Prepares the Twig UI templates directory.
   *
   * @return mixed
   *   TRUE if successful.
   *   A string if not successful.
   */
  public static function prepareTemplatesDirectory() {
    $file_system = \Drupal::service('file_system');

    // Attempt to prepare the templates directory.
    $directory_path = self::DIRECTORY_PATH;
    if (!$file_system->prepareDirectory($directory_path, FileSystem::CREATE_DIRECTORY)) {
      return 'Unable to create templates directory';
    }

    // Write .htaccess file to the templates directory.
    // Delete existing .htaccess file first.
    $file_system->delete($directory_path . '/.htaccess');
    if (!\Drupal::service('file.htaccess_writer')->write($directory_path)) {
      return 'Unable to create .htaccess file in templates directory';
    }

    return TRUE;
  }

}
